///////////////////////////////////////////////////////////////////////////////
//
// System iNtrusion Analysis and Reporting Environment (SNARE) - Audit Daemon
//
// Audit Daemon
//
////////////////////////////////////////////////////////////////////////////////
// Copyright 1999-2004 InterSect Alliance Pty Ltd
// - http://www.intersectalliance.com/
//
// History:
//   1999-2004: Prehistory
//   20040220:  Merged many of Mark & Jons changes.
//              Started integrating micro embedded web server from Solaris code.
//
////////////////////////////////////////////////////////////////////////////////
//	Compile: gcc -g -O6 -o auditd auditd.c webserver.c WebPages.c
////////////////////////////////////////////////////////////////////////////////


#include <errno.h>		// perror()
#include <fcntl.h>
#include <signal.h>
#include <stdint.h>		// uint32_t
#include <stdarg.h>		// varargs
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>		// read(), pause(), unlink()
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <regex.h>		// Regular expression matches.

#include <netdb.h>		// gethostbyname
#include <sys/socket.h>

#include <sys/mman.h>		// mmap()
#include <sys/stat.h>		// file modes
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <linux/kdev_t.h>	// MAJOR and MINOR macros
//#include <linux/limits.h>	// PATH_MAX
#include <limits.h>
//#include <linux/net.h>		// AUDIT_CONNECT and AUDIT_ACCEPT

//#include <linux/saudit.h>

#include "auditd.h"
#include "webserver.h"

extern int InitWebServer(unsigned short, char *, char *);
extern int StartThread();


//////////////////////////////////

Node *head;
const char device_name[] = AUDITDEV_FILE;	// device name
int device_fd = -1;			// File descriptor for /proc/snare
int process_id;				// Daemon's process id
int AuditMode = AUDIT_BY_OBJECTIVE;	// By Objective, or By Event.
int AuditDestination = AUDIT_TO_STDOUT;	// Send to STDOUT unless otherwise stated.
int SyslogDestination=13;
FILE *AuditFile = (FILE *) NULL;
int AuditSocket = 0;
struct sockaddr_in AuditSocketName;
char hostid[MAX_HOSTID] = "";	// Host name - hopefully fully qualified
int signal_flags = 0x00;

// Global variables for the web server
int webpid=0;
int remote_allow=0;
int remote_webport=6161;
char remote_restrictip[16]="";
char remote_password[256]="";

////////////////////////////////////////////////////////////////////////////////
// Set everything up, then wait for data to write out
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
	struct sigaction act;	// For sigaction()
	header_token header;
	int done = 0;

	printf("SNARE audit daemon: version %d.%d%d starting up\n\
Inter[01;31mS[0mect Alliance Pty Ltd\n\
[01;34mhttp://www.intersectalliance.com/[0m\n", SNAREAUDIT_MAJOR_VERSION, SNAREAUDIT_MINOR_VERSION, SNAREAUDIT_PATCH_VERSION);

	// become daemon, die if it fails
	// Only do this if we are not sending data to stdout.

//      if((pid = fork()) < 0)
//              error_sys("auditd: can't become daemon, exiting \n");
//      else if(pid) // I'm the parent, suicide
//              exit(0);
//      start a new session

	process_id = setsid();

	// Read configuration file.
	if ((device_fd = open(device_name, O_RDONLY, 0)) < 0) {
		// There's a chance that snare is resetting. Wait for a bit for the file to be released.
		sleep(5);
		if ((device_fd = open(device_name, O_RDONLY, 0)) < 0) {
			abort_audit("SNARE audit daemon: I cannot connect to the audit capability in \
the kernel via the file /proc/snare. Please verify that you are running a kernel \
that has Snare integrated, and only one copy of the audit daemon is running.", 0);
		}
	}
	// Initialise our linked list.

	CreateLinkedList();

	// First, flush existing audit events that may have been left over
	// From a previous auditd incarnation

	if (ioctl(device_fd, AUDIT_FLUSH, 0) < 0) {
		abort_audit
		    ("audit daemon: can't flush audit events. Exiting.",
		     0);
	}
	// Here: read config file, set audit events, etc.
	// Auditmode is either by event, or by objective

	AuditMode = read_config_file();

	// XXX read_config_file actually never returns -1 at present

	if (AuditMode == -1) {
		abort_audit("auditd: can't parse config file, exiting", 0);
	}

	// Fill in the hostid with the fully qualified domain name,
	// if it hasn't already been grabbed from the configuration file.

	if (!hostid || !*hostid) {
		getfqdn(hostid);
	}

	// set up signal handling
	act.sa_handler = sig_handler;
	sigfillset(&(act.sa_mask));	// block all signals while handling
	act.sa_flags = 0;		// no flags

	if (sigaction(SIGTERM, &act, NULL) || sigaction(SIGHUP, &act, NULL)
	    || sigaction(SIGINT, &act, NULL)
	    || sigaction(SIGQUIT, &act, NULL)
	    || sigaction(SIGALRM, &act, NULL)
	    || sigaction(SIGTSTP, &act, NULL)
	    || sigaction(SIGUSR1, &act, NULL)
	    || sigaction(SIGCONT, &act, NULL)) {
		abort_audit("auditd: problem setting signals, exiting", 0);
	}

	// Do we need to start up our embedded web server?
	if(remote_allow) {
		webpid=fork();
		if(webpid==0) {
			int ret=0;
			// Child process - Run our web server
			while(InitWebServer(remote_webport,remote_restrictip,remote_password) == -1) {
				// Problem binding. Sleep for a moment.
				sleep(5);
			}
			while((ret = StartThread()) == 1) {
				// Problem binding. Keep trying.
				sleep(5);
			}

			exit(0);
		}
	}



	printf("SNARE audit daemon: driver open, starting audit\n");

	if (ioctl(device_fd, AUDIT_START, 0) < 0) {
		abort_audit("auditd: can't start auditing, exiting", 0);
	}

	// Read the header data from the device, and determine
	// the size of the remaining record.
	while (!done) {

		done = safe_read(device_fd, &header, sizeof(header_token));

		if (!done) {
			switch (header.event_class) {
			case AUDIT_CLASS_IO:
				process_io_class(&header);
				break;

			case AUDIT_CLASS_EXEC:
				process_ex_class(&header);
				break;

			case AUDIT_CLASS_PC:
				process_pc_class(&header);
				break;

			case AUDIT_CLASS_CH:
				process_ch_class(&header);
				break;

			case AUDIT_CLASS_CP:
				process_cp_class(&header);
				break;

			case AUDIT_CLASS_SU:
				process_su_class(&header);
				break;

			case AUDIT_CLASS_NET:
				process_nt_class(&header);
				break;

			case AUDIT_CLASS_AD:
				process_ad_class(&header);
				break;

			default:
				if(header.event_class > 0) {
					printf
					    ("Auditd: Problem - Data read by daemon is not a recognised audit class - %d\n",
					     header.event_class);
				}

				// Note: Only read to a max of 4096 here. There is an outside chance that
				// something may have been corrupted, and I don't want any overflows.

				if (header.event_size > MAX_AUDITREC) {
					header.event_size = MAX_AUDITREC;
				}

			}
		}

		if (signal_flags) {
			sync_signal_handler();
		}
	}

	// if we're here, something went very wrong
	abort_audit("Audit Daemon has terminated unexpectedly.", 0);
	return(1);
}

////////////////////////////////////////////////////////////////////////////////
// Catch signals sent by the audit kernel capability
////////////////////////////////////////////////////////////////////////////////

void sig_handler(int signal_id)
{
	switch (signal_id) {
	case SIGTERM:
	case SIGALRM:
	case SIGHUP:
	case SIGINT:
	case SIGQUIT:
		signal_flags |= AUDIT_SIGNAL_STOP;
		break;

	case SIGTSTP:
		signal_flags |= AUDIT_SIGNAL_SUSPEND;
		break;

	case SIGCONT:
		signal_flags |= AUDIT_SIGNAL_RESUME;
		break;

	case SIGUSR1:
		// Could use HUP here instead/as well...
		signal_flags |= AUDIT_SIGNAL_REFRESH;
		break;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function to safely (synchronously) handle signal flags within our event loop
////////////////////////////////////////////////////////////////////////////////
void sync_signal_handler()
{
	if (signal_flags & AUDIT_SIGNAL_STOP) {
		abort_audit("Audit daemon: finished.", 0);
	}

	if (signal_flags & AUDIT_SIGNAL_SUSPEND) {
		// CTRL-Z received. Tell the audit daemon to stop sending data.
		// IGNORE suspend signals.
		// ioctl(device_fd, AUDIT_STOP, 0);

		signal_flags &= ~AUDIT_SIGNAL_SUSPEND;
	}

	if (signal_flags & AUDIT_SIGNAL_RESUME) {
		// And continue signals.
		// ioctl(device_fd, AUDIT_START, 0);

		signal_flags &= ~AUDIT_SIGNAL_RESUME;
	}

	if (signal_flags & AUDIT_SIGNAL_REFRESH) {
		// Re-read the configuration file.
		// First, flush existing audit events.

		if (ioctl(device_fd, AUDIT_FLUSH, 0) < 0) {
			abort_audit
			    ("audit daemon: can't flush audit events during config reload. Exiting.",
			     0);
		}

		// Free the old linked list.

		DestroyList();
		CreateLinkedList();
		strcpy(hostid, "");

		close_outputs();

		// Reset the event count and the event lost count.

		AuditMode = read_config_file();

		// XXX read_config_file actually never returns -1 at present

		if (AuditMode == -1) {
			abort_audit
			    ("auditd: can't parse config file during reload, exiting",
			     0);
		}
		// Fill in the hostid with the fully qualified domain name,
		// if it hasn't already been grabbed from the configuration file.

		if (!hostid || !*hostid) {
			getfqdn(hostid);
		}

		signal_flags &= ~AUDIT_SIGNAL_REFRESH;
	}
}

int turn_event_on(int eventnumber)
{
	return (ioctl(device_fd, AUDIT_EVENT_ON, eventnumber));
}

void abort_audit(char *msg, int err)
{
	// perror first so we don't overwrite errno with later errors

	if (err) {
		perror(msg);
	} else {
		printf("%s\n",msg);
	}

	if(webpid) {
		printf("Terminating Web Server\n");

		// Ask the web server to terminate nicely if it is open.
		kill(webpid,SIGUSR1);

		waitpid(webpid,NULL,0);
	}

	if (device_fd >= 0) {
		ioctl(device_fd, AUDIT_STOP, 0);	// stop kernel auditing
		close(device_fd);
		device_fd = -1;
	}

	close_outputs();
	DestroyList();

	exit(err);
}

// read function that doesn't break if it receives a signal.  returns
// 0 on success
int safe_read(int fd, void *ptr, int size)
{
	int readres = 0;
	int total = 0;
	while (total < size && !signal_flags) {
		readres = read(fd, ptr+total, size-total);

		if (readres > 0) {
			total += readres;
		} else {
			if(readres == 0) {
				return(-1);
			} else if (errno != EINTR) {
				return errno;	// failure
			}
		}
	}

	return 0;
}

void close_outputs()
{
	if (AuditDestination & AUDIT_TO_FILE) {
		if (AuditFile != (FILE *) NULL) {
			fclose(AuditFile);
		}
	}

	if (AuditDestination & AUDIT_TO_NETWORK) {
		if (AuditSocket) {
			close(AuditSocket);
		}
	}

	fflush(stdout);
}

/////////////////////////////////////////////////////////////////////////////
// Process events that relate to input/output
/////////////////////////////////////////////////////////////////////////////
int process_io_class(header_token * header)
{
	io_class io_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC] = "";	// Event data storage.
	char tempdata[MAX_AUDITREC] = "";	// for temporary storage if required.
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char pathstore[PATH_MAX];
	char userstore[MAX_USERNAME];
	char options[MAX_OPTIONS] = "";
	char timestr[TIMEBUF] = "";
	int delim = 0;		// temporary variable to indicate whether a | is required.
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 1;

	io_event.t_header = *header;

	// Note that this assumes that the return_token will ALWAYS be the
	// second element in each structure.
	if (safe_read
	    (device_fd, (void *) &io_event.t_header + sizeof(header_token),
	     io_event.t_header.event_size)) {
		return 0;
	}

	// Do a quick look up and resolution of the user and event path so
	// we can do an early objective check
	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	resolved_path(io_event.t_pwd.path, io_event.t_path.path,
		      pathstore);

	// all events we are looking at in the io_class take returncode>=0
	// to be success
	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	// If we're doing an open, we'll need to check on the options to
	// pass to our objective sweep, so go ahead and get those figured
	// out here
	if (header->event_id == AUDIT_open) {
		calc_open_options(options, &io_event);
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		// Now let's take a quick peek for objectives.  If we don't find
		// any, we're done with this event
		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  pathstore,
						  options,
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	// now we know we need to actually figure out what the event (and
	// possibly objective) should look like
	strncpy(eventdata,"event",sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (io_event.t_header.event_id) {

	case AUDIT_open:

		// we precalculated options

		va_eventcat(eventdata, 3, "open(", options, ")");

		if (do_objectives) {
			if (header->returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s has been opened ");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to open the file %s ");
			}

			// we'll already have most of this stuff in options, but if
			// we're in objectives mode, get the human readable strings

			// GNU allows O_RDONLY|O_WRONLY as a valid alternative to O_RDWR

			if (io_event.t_attributes.mode & O_WRONLY
			    && (!(io_event.t_attributes.mode & O_RDONLY)))
			{
				eventcat(objectivetext, "(write-only)");
			} else if (io_event.t_attributes.mode & O_RDWR ||
				   ((io_event.t_attributes.mode & O_WRONLY)
				    && (io_event.t_attributes.
					mode & O_RDONLY))) {
				eventcat(objectivetext,
					 "(for both read and write)");
			} else {
				eventcat(objectivetext, "(read only)");
			}

			eventcat(objectivetext, " by the user %s");
		}

		break;

	case AUDIT_mkdir:

		eventcat(eventdata, "mkdir()");

		if (do_objectives) {
			if (header->returncode >= 0) {
				eventcat(objectivetext,
					 "The directory %s has been created by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to create the directory %s by the user %s");
			}
		}

		break;

	case AUDIT_unlink:

		eventcat(eventdata, "unlink()");

		if (do_objectives) {
			if (header->returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s has been removed by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to remove the file %s by the user %s");
			}
		}

		break;

	case AUDIT_mknod:

		eventcat(eventdata, "mknod(");

		if (io_event.t_attributes.mode & S_IFREG) {
			eventcat(eventdata, "S_IFREG");
			delim = 1;
		}

		if (io_event.t_attributes.mode & S_IFCHR) {
			if (delim) {
				eventcat(eventdata, "|S_IFCHR");
			} else {
				eventcat(eventdata, "S_IFCHR");
				delim = 1;
			}
		}

		if (io_event.t_attributes.mode & S_IFBLK) {
			if (delim) {
				eventcat(eventdata, "|S_IFBLK");
			} else {
				eventcat(eventdata, "S_IFBLK");
				delim = 1;
			}
		}

		if (io_event.t_attributes.mode & S_IFIFO) {
			if (delim) {
				eventcat(eventdata, "|S_IFIFO");
			} else {
				eventcat(eventdata, "S_IFIFO");
				delim = 1;
			}
		}
		// Otherwise, it is a normal file.

		eventcat(eventdata, ")");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The special file %s has been created by the user %s using mknod");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to create the special file %s by the user %s using mknod");
			}
		}

		break;

	case AUDIT_rmdir:

		eventcat(eventdata, "rmdir()");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The directory %s has been removed by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to remove the directory %s by the user %s");
			}
		}

		break;

	case AUDIT_chmod:

		eventcat(eventdata, "chmod()");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The permissions associated with the file %s has been changed by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the permissions of the file %s by the user %s");
			}
		}

		break;

	case AUDIT_truncate:

		eventcat(eventdata, "truncate()");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s was truncated by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to truncate the file %s by the user %s");
			}
		}

		break;

	case AUDIT_chroot:

		eventcat(eventdata, "chroot()");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The directory %s has now been set to the system root by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to set the system root to the directory %s by the user %s");
			}
		}

		break;

	case AUDIT_ftruncate:

		eventcat(eventdata, "ftruncate()");

		if (do_objectives) {
			if (io_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s was truncated by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to truncate the file %s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time
	ctime_r(&io_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	// Only problem with the stuff below is that for each file open (for example),
	// another 4 file opens will be generated as the getpwuid/getgrgid tasks
	// interrogate the password file. Note that events generated by the audit daemon are ignored though.
	// Still, I'll try to keep these calls to a minimum.

	add_user_token(eventdata, &(io_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(io_event.t_header));

	va_eventcat(eventdata, 5, DELIMITER, "path", IDELIMITER, pathstore,
		    DELIMITER);

	// NOTE: this is the requested file write mode.
	// It will be modified by the processes UMASK
	// Thus, the permissions shown may not represent the final file.
	// open events - only available if file created
	if ((io_event.t_header.event_id == AUDIT_open &&
	     (io_event.t_attributes.mode & O_CREAT)) ||
	    io_event.t_header.event_id == AUDIT_mkdir) {
		va_eventcat(eventdata, 4,
			    "attributes",
			    IDELIMITER,
			    textmode(io_event.t_attributes.createmode,
				     tempdata, sizeof(tempdata)),
			    DELIMITER);

	} else if (io_event.t_header.event_id == AUDIT_mknod &&
		   (io_event.t_attributes.mode & S_IFBLK ||
		    io_event.t_attributes.mode & S_IFCHR)) {
		// Mknod

		snprintf(tempdata, MAX_AUDITREC, "%d %d",
			 (unsigned short) MAJOR(io_event.t_attributes.
						createmode),
			 (unsigned short) MINOR(io_event.t_attributes.
						createmode));

		va_eventcat(eventdata, 4,
			    "attributes", IDELIMITER, tempdata, DELIMITER);
	} else if (io_event.t_header.event_id == AUDIT_chmod) {
		va_eventcat(eventdata, 4,
			    "attributes",
			    IDELIMITER,
			    textmode(io_event.t_attributes.createmode,
				     tempdata, sizeof(tempdata)),
			    DELIMITER);
	} else if (io_event.t_header.event_id == AUDIT_truncate
		   || io_event.t_header.event_id == AUDIT_ftruncate) {
		snprintf(tempdata, MAX_AUDITREC, "%ld",
			 io_event.t_attributes.createmode);

		va_eventcat(eventdata, 4,
			    "attributes", IDELIMITER, tempdata, DELIMITER);
	}
	// Return code

	add_return_token(eventdata, &(io_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objective str and log
	// everything

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, pathstore, userstore);

		ProcessObjectives(objectiveNode,
				  &io_event.t_header,
				  pathstore,
				  options,
				  objectivestr, eventdata, timestr,
				  ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}

void calc_open_options(char *options, io_class * io_event)
{
	// GNU allows O_RDONLY|O_WRONLY as a valid alternative to O_RDWR

	options[0] = '\0';

	if (io_event->t_attributes.mode & O_WRONLY
	    && (!(io_event->t_attributes.mode & O_RDONLY))) {
		eventcat(options, "O_WRONLY");
	} else if (io_event->t_attributes.mode & O_RDWR ||
		   ((io_event->t_attributes.mode & O_WRONLY) &&
		    (io_event->t_attributes.mode & O_RDONLY))) {
		eventcat(options, "O_RDWR");
	} else {
		eventcat(options, "O_RDONLY");
	}

	// Now for all the additional flags

	if (io_event->t_attributes.mode & O_CREAT) {
		eventcat(options, "|O_CREAT");
	}

	if (io_event->t_attributes.mode & O_EXCL) {
		eventcat(options, "|O_EXCL");
	}

	if (io_event->t_attributes.mode & O_NOCTTY) {
		eventcat(options, "|O_NOCTTY");
	}

	if (io_event->t_attributes.mode & O_TRUNC) {
		eventcat(options, "|O_TRUNC");
	}

	if (io_event->t_attributes.mode & O_APPEND) {
		eventcat(options, "|O_APPEND");
	}

	if (io_event->t_attributes.mode & O_NONBLOCK) {
		eventcat(options, "|O_NONBLOCK");
	}

	if (io_event->t_attributes.mode & O_SYNC) {
		eventcat(options, "|O_SYNC");
	}
#ifdef __USE_GNU
	if (io_event->t_attributes.mode & O_NOFOLLOW) {
		eventcat(options, "|O_NOFOLLOW");
	}

	if (io_event->t_attributes.mode & O_DIRECTORY) {
		eventcat(options, "|O_DIRECTORY");
	}
#endif

#ifdef __USE_LARGEFILE64
	if (io_event->t_attributes.mode & O_LARGEFILE) {
		eventcat(options, "|O_LARGEFILE");
	}
#endif
}


////////////////////////////////////////////////////////////////////////////////
// Process events that relate to input/output CHMODs
////////////////////////////////////////////////////////////////////////////////
int process_ch_class(header_token * header)
{
	ch_class ch_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char tempdata[MAX_AUDITREC];	// for temporary storage if required.
	char tempdata2[MAX_AUDITREC];	// for temporary storage if required.
	char pathstore[PATH_MAX];
	char userstore[MAX_USERNAME];
	char timestr[TIMEBUF] = "";
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	ch_event.t_header = *header;

	// Note that this assumes that the path_token will ALWAYS be the second element in each structure.

	if (safe_read
	    (device_fd, (void *) &ch_event.t_header + sizeof(header_token),
	     ch_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user and event path so
	// we can do an early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));
	resolved_path(ch_event.t_pwd.path, ch_event.t_path.path,
		      pathstore);

	// all events we are looking at in the ch_class take returncode>=0
	// to be success

	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		// Now let's take a quick peek for objectives.  If we don't find
		// any, we're done with this event

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  pathstore,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (ch_event.t_header.event_id) {

	case AUDIT_chown:

		eventcat(eventdata, "chown()");

		if (do_objectives) {
			if (ch_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The owner associated with the file %s has been changed by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the owner of the file %s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&ch_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(ch_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(ch_event.t_header));

	// Use resolved_path to work out the correct path for the file.

	va_eventcat(eventdata, 5,
		    DELIMITER, "path", IDELIMITER, pathstore, DELIMITER);

	if (ch_event.t_header.event_id == AUDIT_chown) {
		// owner

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 ch_event.t_owner.owner);
		snprintf(tempdata2, MAX_AUDITREC, "%d",
			 (gid_t) ch_event.t_owner.group);

		va_eventcat(eventdata, 12, "owner", IDELIMITER, passname(ch_event.t_owner.owner), "(", tempdata,	// uid
			    ")", IDELIMITER, groupname(ch_event.t_owner.group), "(", tempdata2,	// gid
			    ")", DELIMITER);
	}
	// Return code
	add_return_token(eventdata, &(ch_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objective str and log
	// everything

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, pathstore, userstore);

		ProcessObjectives(objectiveNode,
				  &ch_event.t_header,
				  pathstore,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}


////////////////////////////////////////////////////////////////////////////////
// Process execution events
////////////////////////////////////////////////////////////////////////////////
int process_ex_class(header_token * header)
{
	ex_class ex_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char tempdata[MAX_AUDITREC];	// for temporary storage if required.
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char pathstore[PATH_MAX];
	char userstore[MAX_USERNAME];
	char timestr[TIMEBUF] = "";
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	ex_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &ex_event.t_header + sizeof(header_token),
	     ex_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user and event path so
	// we can do an early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));
	resolved_path(ex_event.t_pwd.path, ex_event.t_path.path,
		      pathstore);

	// all events we are looking at in the ex_class take returncode>=0
	// to be success

	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		// Now let's take a quick peek for objectives.  If we don't find
		// any, we're done with this event

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  pathstore,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}


	switch (ex_event.t_header.event_id) {

	case AUDIT_execve:
		eventcat(eventdata, "execve()");

		if (do_objectives) {
			if (ex_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The program %s has been executed by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to execute the program %s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&ex_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(ex_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(ex_event.t_header));

	// Use resolved_path to work out the correct path for the file.
	va_eventcat(eventdata, 9,
		    DELIMITER,
		    "path",
		    IDELIMITER,
		    resolved_path(ex_event.t_pwd.path,
				  ex_event.t_path.path, tempdata),
		    DELIMITER, "arguments", IDELIMITER,
		    strip(ex_event.t_execargs.args), DELIMITER);

	add_return_token(eventdata, &(ex_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objectivestr and log
	// everything

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, pathstore, userstore);

		ProcessObjectives(objectiveNode,
				  &ex_event.t_header,
				  pathstore,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}
	return (0);
}

////////////////////////////////////////////////////////////////////////////////
// Process process-control events.
////////////////////////////////////////////////////////////////////////////////
int process_pc_class(header_token * header)
{
	pc_class pc_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char userstore[MAX_USERNAME];
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char timestr[TIMEBUF] = "";
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	pc_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &pc_event.t_header + sizeof(header_token),
	     pc_event.t_header.event_size)) {
		return 0;
	}

	// Do a quick look up and resolution of the user so we can do an
	// early objective check
	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	// all events we are looking at in the pc_class take returncode>=0
	// to be success
	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  header->processname,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (pc_event.t_header.event_id) {

	case AUDIT_exit:

		eventcat(eventdata, "exit()");
		// Note: exit does not return.. put this in for completeness though.

		if (do_objectives) {
			if (pc_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s, owned by the user %s, has exited");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to exit the process %s by the user %s");
			}
		}

		break;

	case AUDIT_fork:

		eventcat(eventdata, "fork()");

		if (do_objectives) {
			if (pc_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s, owned by the user %s, has forked off a child process");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to fork a process %s by the user %s");
			}
		}

		break;

	case AUDIT_reboot:

		eventcat(eventdata, "reboot()");

		if (do_objectives) {
			if (pc_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s, owned by the user %s, has issued a reboot command");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to reboot the system from the process %s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&pc_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(pc_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(pc_event.t_header));
	eventcat(eventdata, DELIMITER);
	add_return_token(eventdata, &(pc_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objectivestr and log
	// everything
	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, pc_event.t_header.processname,
			 userstore);

		ProcessObjectives(objectiveNode,
				  &pc_event.t_header,
				  pc_event.t_header.processname,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}

////////////////////////////////////////////////////////////////////////////////
// Process network related events.
////////////////////////////////////////////////////////////////////////////////
int process_nt_class(header_token * header)
{
	nt_class nt_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char tempdata[MAX_AUDITREC];	// for temporary storage if required.
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char userstore[MAX_USERNAME];
	char options[MAX_OPTIONS] = "";
	char timestr[TIMEBUF] = "";
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	nt_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &nt_event.t_header + sizeof(header_token),
	     nt_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user so we can do an
	// early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	// all events we are looking at in the nt_class take returncode>=0
	// to be success

	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		// XXX Note that we are currently not calculating any options
		// for any events in this class.. if we do add options here,
		// we'll need to construct the options string before this point.
		//
		// cf. process_io_class

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  header->processname,
						  options,
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (nt_event.t_header.event_id) {

	case AUDIT_connect:

		eventcat(eventdata, "connect()");

		if (do_objectives) {
			if (nt_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s, owned by the user %s, has connected to ");
				snprintf(tempdata, MAX_AUDITREC,
					 "%s port %d on remote system %s",
					 (nt_event.t_connection.protocol ==
					  IPPROTO_UDP ? "udp" : "tcp"),
					 nt_event.t_connection.dst_port,
					 nt_event.t_connection.dst_ip);
				eventcat(objectivetext, tempdata);
			} else {
				eventcat(objectivetext,
					 "A network connect failed from the process %s by the user %s to ");
				snprintf(tempdata, MAX_AUDITREC,
					 "%s port %d on %s",
					 (nt_event.t_connection.protocol ==
					  IPPROTO_UDP ? "udp" : "tcp"),
					 nt_event.t_connection.dst_port,
					 nt_event.t_connection.dst_ip);
				eventcat(objectivetext, tempdata);
			}
		}

		break;

	case AUDIT_accept:

		eventcat(eventdata, "accept()");

		if (do_objectives) {
			if (nt_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s, owned by the user %s, has accepted a connection to ");
				snprintf(tempdata, MAX_AUDITREC,
					 "%s port %d from remote system %s",
					 (nt_event.t_connection.protocol ==
					  IPPROTO_UDP ? "udp" : "tcp"),
					 nt_event.t_connection.dst_port,
					 nt_event.t_connection.src_ip);
				eventcat(objectivetext, tempdata);
			} else {
				eventcat(objectivetext,
					 "A network accept failed from the process %s by the user %s to ");
				snprintf(tempdata, MAX_AUDITREC,
					 "%s port %d on %s",
					 (nt_event.t_connection.protocol ==
					  IPPROTO_UDP ? "udp" : "tcp"),
					 nt_event.t_connection.dst_port,
					 nt_event.t_connection.src_ip);
				eventcat(objectivetext, tempdata);
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&nt_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(nt_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(nt_event.t_header));

	snprintf(tempdata, MAX_AUDITREC, "%s%s%s%s%d%s%d",
		 nt_event.t_connection.src_ip, IDELIMITER,
		 nt_event.t_connection.dst_ip, IDELIMITER,
		 nt_event.t_connection.src_port, IDELIMITER,
		 nt_event.t_connection.dst_port);

	va_eventcat(eventdata, 5, DELIMITER, "socket", IDELIMITER,
		    tempdata, DELIMITER);

	add_return_token(eventdata, &(nt_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, nt_event.t_header.processname,
			 userstore);

		// the match term is the command that performed the network function.             
		// Options: remote ip address? port?

		ProcessObjectives(objectiveNode,
				  &nt_event.t_header,
				  nt_event.t_header.processname,
				  options,
				  objectivestr, eventdata, timestr,
				  ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}


////////////////////////////////////////////////////////////////////////////
// Process events that relate to file duplications
////////////////////////////////////////////////////////////////////////////
int process_cp_class(header_token * header)
{
	cp_class cp_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char spathstore[PATH_MAX];
	char dpathstore[PATH_MAX];
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char userstore[MAX_USERNAME];
	char timestr[TIMEBUF] = "";
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;
	int finished_source_check = 0;

	cp_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &cp_event.t_header + sizeof(header_token),
	     cp_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user and the source and
	// destination paths so we can do an early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	resolved_path(cp_event.t_pwd.path, cp_event.t_sourcepath.path,
		      spathstore);
	resolved_path(cp_event.t_pwd.path, cp_event.t_destpath.path,
		      dpathstore);

	// all events we are looking at in the cp_class take returncode>=0
	// to be success

	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		// we've got two separate checks to do here.. note that this
		// logic is mated with that at the end of this func

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  spathstore,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			finished_source_check = 1;

			objectiveNode = FindNextObjective(head,
							  header->event_id,
							  userstore,
							  dpathstore,
							  "",
							  header->
							  returncode,
							  ObjectiveReturn);
		}

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (cp_event.t_header.event_id) {

	case AUDIT_symlink:

		eventcat(eventdata, "symlink()");

		if (do_objectives) {
			if (cp_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s has been symbolically linked to the file %s by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to symbolically link the file %s to the file %s by the user %s");
			}
		}

		break;

	case AUDIT_link:

		eventcat(eventdata, "link()");

		if (do_objectives) {
			if (cp_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s has been linked to the file %s by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to link the file %s to the file %s by the user %s");
			}
		}

		break;

	case AUDIT_rename:

		eventcat(eventdata, "rename()");

		if (do_objectives) {
			if (cp_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The file %s has been renamed to %s by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to rename the file %s to %s by the user %s");
			}
		}

		break;

	case AUDIT_mount:

		eventcat(eventdata, "mount()");

		if (do_objectives) {
			if (cp_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The directory %s has been mounted at the path %s by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to mount the directory %s at the path %s by the user %s");
			}
		}

		break;

	case AUDIT_umount:

		// TODO
		// May change umount back to a different class, since it only has a single path entry.
		eventcat(eventdata, "umount()");

		if (do_objectives) {
			if (cp_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The directory %s has been unmounted%s by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to unmount the directory %s%s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&cp_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(cp_event.t_header));
	eventcat(eventdata, DELIMITER);

	// Process
	add_process_token(eventdata, &(cp_event.t_header));

	va_eventcat(eventdata, 9,
		    DELIMITER,
		    "path",
		    IDELIMITER,
		    spathstore,
		    DELIMITER,
		    "destpath", IDELIMITER, dpathstore, DELIMITER);

	add_return_token(eventdata, &(cp_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objective str and log
	// everything

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, spathstore, dpathstore, userstore);

		// first check source path
		if (!finished_source_check) {
			ProcessObjectives(objectiveNode,
					  &cp_event.t_header,
					  spathstore,
					  "",
					  objectivestr,
					  eventdata,
					  timestr, ObjectiveReturn);

			// now start searching for a dpathstore match

			objectiveNode = FindNextObjective(head,
							  header->event_id,
							  userstore,
							  dpathstore,
							  "",
							  header->
							  returncode,
							  ObjectiveReturn);
		}

		// second check dest path
		ProcessObjectives(objectiveNode,
				  &cp_event.t_header,
				  dpathstore,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}

////////////////////////////////////////////////////////////////////////////////
// Process setuid/setgid
////////////////////////////////////////////////////////////////////////////////
int process_su_class(header_token * header)
{
	su_class su_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char tempdata[MAX_AUDITREC];	// for temporary storage if required.
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char userstore[MAX_USERNAME];
	char timestr[TIMEBUF] = "";
	int TargetID = TARGET_EID;
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	su_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &su_event.t_header + sizeof(header_token),
	     su_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user so we can do an
	// early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	// all events we are looking at in the su_class take returncode>=0
	// to be success

	if (header->returncode >= 0) {
		ObjectiveReturn = RETURNCODE_SUCCESS;
	} else {
		ObjectiveReturn = RETURNCODE_FAILURE;
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  header->processname,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata, "event", sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (su_event.t_header.event_id) {

	case AUDIT_setuid:

		eventcat(eventdata, "setuid()");
		TargetID = TARGET_EID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed effective userid from the user %s to %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the effective user id of the process %s by the user %s to %s");
			}
		}

		break;

	case AUDIT_setreuid:

		eventcat(eventdata, "setreuid()");
		TargetID = TARGET_EID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed real and effective user id from the user %s to %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the real and effective user id of the process %s by the user %s to the user %s");
			}
		}

		break;

	case AUDIT_setresuid:

		eventcat(eventdata, "setresuid()");
		TargetID = TARGET_EID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed real, effective and saved user id from the user %s to %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the real, effective and saved user id of the process %s by the user %s to the user %s");
			}
		}

		break;

	case AUDIT_setgid:

		eventcat(eventdata, "setgid()");
		TargetID = TARGET_GID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed effective group id, called by the user %s, to the group %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the effective group id of the process %s by the user %s to the group %s");
			}
		}

		break;

	case AUDIT_setregid:

		eventcat(eventdata, "setregid()");
		TargetID = TARGET_GID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed real and effective group id, called by the user %s, to the group %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the real and effective group id of the process %s by the user %s to the group %s");
			}
		}

		break;

	case AUDIT_setresgid:

		eventcat(eventdata, "setresgid()");
		TargetID = TARGET_GID;

		if (do_objectives) {
			if (su_event.t_header.returncode >= 0) {
				eventcat(objectivetext,
					 "The process %s has changed real, effective and saved group id, called by the user %s, to the group %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to change the real, effective and saved group id of the process %s by the user %s to the group %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&su_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(su_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(su_event.t_header));
	eventcat(eventdata, DELIMITER);

	if (su_event.t_header.event_id == AUDIT_setuid
	    || su_event.t_header.event_id == AUDIT_setreuid
	    || su_event.t_header.event_id == AUDIT_setresuid) {
		// Target user

		va_eventcat(eventdata, 2, "target", IDELIMITER);

		if (su_event.t_target.id != -1) {
			eventcat(eventdata,
				 passname(su_event.t_target.id));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.id);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	} else if (su_event.t_header.event_id == AUDIT_setgid
		   || su_event.t_header.event_id == AUDIT_setregid
		   || su_event.t_header.event_id == AUDIT_setresgid) {
		// Target user
		va_eventcat(eventdata, 2, "target", IDELIMITER);

		if (su_event.t_target.id != -1) {
			eventcat(eventdata,
				 groupname(su_event.t_target.id));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.id);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	}

	if (su_event.t_header.event_id == AUDIT_setreuid
	    || su_event.t_header.event_id == AUDIT_setresuid) {
		// Target user real id
		eventcat(eventdata, IDELIMITER);

		if (su_event.t_target.rid != -1) {
			eventcat(eventdata,
				 passname(su_event.t_target.rid));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.rid);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	} else if (su_event.t_header.event_id == AUDIT_setregid
		   || su_event.t_header.event_id == AUDIT_setresgid) {
		// Target user real id
		eventcat(eventdata, IDELIMITER);

		if (su_event.t_target.rid != -1) {
			eventcat(eventdata,
				 groupname(su_event.t_target.rid));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.rid);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	}

	if (su_event.t_header.event_id == AUDIT_setresuid) {
		// Target user saved id
		eventcat(eventdata, IDELIMITER);

		if (su_event.t_target.sid != -1) {
			eventcat(eventdata,
				 passname(su_event.t_target.sid));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.sid);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	} else if (su_event.t_header.event_id == AUDIT_setresgid) {
		// Target user saved id
		eventcat(eventdata, IDELIMITER);

		if (su_event.t_target.sid != -1) {
			eventcat(eventdata,
				 groupname(su_event.t_target.sid));
		}

		snprintf(tempdata, MAX_AUDITREC, "%d",
			 su_event.t_target.sid);

		va_eventcat(eventdata, 3, "(", tempdata, ")");
	}

	eventcat(eventdata, DELIMITER);

	add_return_token(eventdata, &(su_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";
		char *target;

		if (TargetID == TARGET_EID) {
			// note!  this is a bit fragile, since we're just getting a
			// pointer to the statically cached username in passname.
			// if we were to interleave two passname() calls between
			// this point and where we use the target pointer, we'd have
			// a problem.  We don't, so we don't.

			target = passname(su_event.t_target.id);
		} else {
			target = groupname(su_event.t_target.id);
		}

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, su_event.t_header.processname,
			 userstore, target);

		ProcessObjectives(objectiveNode,
				  &su_event.t_header,
				  su_event.t_header.processname,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}

////////////////////////////////////////////////////////////////////////////////
// Process execution events
////////////////////////////////////////////////////////////////////////////////
int process_ad_class(header_token * header)
{
	ad_class ad_event;
	Node *objectiveNode = NULL;
	char eventdata[MAX_AUDITREC];	// will probably need to make this bigger.
	char objectivetext[MAX_AUDITREC];	// For the objective text, if required.
	char userstore[MAX_USERNAME];
	char timestr[TIMEBUF] = "";
	int ObjectiveReturn = RETURNCODE_SUCCESS;
	int do_objectives = 0;

	ad_event.t_header = *header;

	if (safe_read
	    (device_fd, (void *) &ad_event.t_header + sizeof(header_token),
	     ad_event.t_header.event_size)) {
		return 0;
	}
	// Do a quick look up and resolution of the user so we can do an
	// early objective check

	strncpy(userstore, passname(header->euser_id), sizeof(userstore));

	// We have to be a bit careful here, because there is are event
	// types in this class which have a special success metric

	if (ad_event.t_header.event_id == AUDIT_create_module) {
		if (ad_event.t_header.returncode < -3) {
			ObjectiveReturn = RETURNCODE_SUCCESS;
		} else {
			ObjectiveReturn = RETURNCODE_FAILURE;
		}
	} else if (ad_event.t_header.event_id == AUDIT_delete_module) {
		if (header->returncode == 0) {
			ObjectiveReturn = RETURNCODE_SUCCESS;
		} else {
			ObjectiveReturn = RETURNCODE_FAILURE;
		}
	}

	if (AuditMode == AUDIT_BY_OBJECTIVE) {
		do_objectives = 1;

		objectiveNode = FindNextObjective(head,
						  header->event_id,
						  userstore,
						  ad_event.t_name.path,
						  "",
						  header->returncode,
						  ObjectiveReturn);

		if ((Node *) NULL == objectiveNode) {
			return (0);
		}
	}

	strncpy(eventdata,"event",sizeof(eventdata));
	eventcat(eventdata, IDELIMITER);	// usually comma.

	if (do_objectives) {
		objectivetext[0] = '\0';
	}

	switch (ad_event.t_header.event_id) {

	case AUDIT_create_module:

		eventcat(eventdata, "create_module()");

		if (do_objectives) {
			if (ad_event.t_header.returncode < -3) {
				eventcat(objectivetext,
					 "The module %s was created by process %s, owned by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to create the module %s from the process %s by the user %s");
			}
		}

		break;

	case AUDIT_delete_module:

		eventcat(eventdata, "delete_module()");

		if (do_objectives) {
			if (ad_event.t_header.returncode == 0) {
				eventcat(objectivetext,
					 "The module %s was deleted by process %s, owned by the user %s");
			} else {
				eventcat(objectivetext,
					 "A failed attempt was made to delete the module %s from the process %s by the user %s");
			}
		}

		break;

	default:
		printf("Unknown event type..\n");
		return (0);
	}

	// Date/time

	ctime_r(&ad_event.t_header.time.tv_sec, timestr);
	timestr[strlen(timestr) - 1] = '\0';

	va_eventcat(eventdata, 3, IDELIMITER, timestr, DELIMITER);

	add_user_token(eventdata, &(ad_event.t_header));
	eventcat(eventdata, DELIMITER);

	add_process_token(eventdata, &(ad_event.t_header));

	// Use resolved_path to work out the correct module name.

	va_eventcat(eventdata, 6,
		    DELIMITER,
		    "path",
		    IDELIMITER, ad_event.t_name.path, DELIMITER,
		    DELIMITER);

	add_return_token(eventdata, &(ad_event.t_header),ObjectiveReturn);
	eventcat(eventdata, DELIMITER);
	add_sequence_token(eventdata);

	// If we are in objective mode, resolve our objective str and log
	// everything

	if (do_objectives) {
		char objectivestr[MAX_AUDITREC] = "";

		snprintf(objectivestr, MAX_AUDITREC,
			 objectivetext, ad_event.t_name.path,
			 ad_event.t_header.processname, userstore);

		ProcessObjectives(objectiveNode,
				  &ad_event.t_header,
				  ad_event.t_name.path,
				  "",
				  objectivestr,
				  eventdata, timestr, ObjectiveReturn);
	} else {
		sendevent(eventdata);
	}

	return (0);
}

// note that add_user_token will spoil the passname and groupname
// caches.  be sure not to retain pointers to passname or groupname
// results past a call to add_user_token.

int add_user_token(char *eventdata, header_token * header)
{
	// static so we don't allocate each call.. lots of them so we can do
	// our token generation in one pass.. 16 bytes because we're only
	// dealing with 32 bit integer values in decimal form

	static char tempdata[16];
	static char tempdata2[16];
	static char tempdata3[16];
	static char tempdata4[16];

	snprintf(tempdata, 16, "%d", header->user_id);
	snprintf(tempdata2, 16, "%d", header->group_id);
	snprintf(tempdata3, 16, "%d", header->euser_id);
	snprintf(tempdata4, 16, "%d", header->egroup_id);

	va_eventcat(eventdata, 21, "user", IDELIMITER, passname(header->user_id),	// user name
		    "(", tempdata,	// UID
		    ")", IDELIMITER, groupname(header->group_id),	// group name
		    "(", tempdata2,	// GID
		    ")", IDELIMITER, passname(header->euser_id),	// effective user name
		    "(", tempdata3,	// EUID
		    ")", IDELIMITER, groupname(header->egroup_id),	// effective group name
		    "(", tempdata4,	// EGID
		    ")");

	return (0);
}

int add_process_token(char *eventdata, header_token * process)
{
	// static so we don't allocate each call
	static char tempdata[MAX_AUDITREC];
	static char tempdata2[MAX_AUDITREC];

	snprintf(tempdata, MAX_AUDITREC, "%d", process->pid);
	snprintf(tempdata2, MAX_AUDITREC, "%d", process->ppid);

	va_eventcat(eventdata, 7, "process", IDELIMITER, tempdata,	// PID
		    IDELIMITER, tempdata2,				// PPID
		    IDELIMITER, process->processname);			// and Name

	return (0);
}

int add_return_token(char *eventdata, header_token * returncode,int ObjectiveReturn)
{
	// static so we don't allocate each call
	static char tempdata[10];
	static char tempdata2[5];

	// Return code
	snprintf(tempdata, 10, "%d", returncode->returncode);
	snprintf(tempdata2, 5, "%d", ObjectiveReturn);

	va_eventcat(eventdata, 5, "return", IDELIMITER, tempdata,
			IDELIMITER, tempdata2);

	return (0);
}

// we use the global tempseq for our scratch space
int add_sequence_token(char *eventdata)
{
	static unsigned int sequence = 1;
	static char tempseq[10];	// Enough for MAXINT (65535)

	// Return code

	snprintf(tempseq, MAX_AUDITREC, "%d", sequence);

	va_eventcat(eventdata, 3, "sequence", IDELIMITER, tempseq);

	sequence++;
	if (sequence >= INT_MAX) {
		sequence = 1;
	}

	return (0);
}

////////////////////////////////////////////////////////////////////////////////
// strcat that respects the maximum size of an audit record
// (<string> buffer is assumed to be of size MAX_AUDITREC)
////////////////////////////////////////////////////////////////////////////////
char *eventcat(char *string, char *append)
{
	char *dest_ptr, *end_orig, *end_marker, *src_ptr;

	if (!string || !append || !*append) {
		return string;
	}

	dest_ptr = string;
	src_ptr = append;

	end_marker = dest_ptr + MAX_AUDITREC-1;
	*end_marker = '\0';	// guard

	while (*dest_ptr) {
		dest_ptr++;
	}

	end_orig = dest_ptr + 1;

	while ((dest_ptr < end_marker) && *src_ptr) {
		*dest_ptr++ = *src_ptr++;
	}

	if (!*src_ptr) {
		*dest_ptr = '\0';
	} else {
		*end_orig = '\0';
	}

	return string;
}

////////////////////////////////////////////////////////////////////////////////
// efficient varargs multi-string strcat with bounds checking
// (<string> buffer is assumed to be of size MAX_AUDITREC)
//
// count must be the number of strings to concatenate onto <string> or else
// bad things will happen.  caveat coder.
//
////////////////////////////////////////////////////////////////////////////////
char *va_eventcat(char *string, int count, ...)
{
	va_list ap;

	int i;

	char *dest_ptr, *end_orig, *end_marker, *src_ptr;

	if (!string) {
		return string;
	}

	dest_ptr = string;
	end_marker = dest_ptr + MAX_AUDITREC -1;
	*end_marker = '\0';	// guard

	while (*dest_ptr) {
		dest_ptr++;
	}

	i = 0;

	va_start(ap, count);

	while (i++ < count) {
		end_orig = dest_ptr + 1;
		src_ptr = va_arg(ap, char *);

		if (!src_ptr) {
			continue;
		}

		while ((dest_ptr < end_marker) && *src_ptr) {
			*dest_ptr++ = *src_ptr++;
		}

		if (!*src_ptr) {
			*dest_ptr = '\0';
		} else {
			*end_orig = '\0';
		}
	}

	va_end(ap);

	return string;
}

// Return a text-type mode
char *textmode(int mode, char *t_mode, int size)
{
	int i;
	// Need at least 9 characters
	if (size < 9)
		return NULL;

	t_mode[0] = '\0';

	        for(i = 2; i >= 0; i--) {
                switch((mode >> (i*3)) & 7) {
                        case 0:
                                strcat(t_mode,"---");break;
                        case 1:
                                strcat(t_mode,"--x");break;
                        case 2:
                                strcat(t_mode,"-w-");break;
                        case 3:
                                strcat(t_mode,"-wx");break;
                        case 4:
                                strcat(t_mode,"r--");break;
                        case 5:
                                strcat(t_mode,"r-x");break;
                        case 6:
                                strcat(t_mode,"rw-");break;
                        case 7:
                                strcat(t_mode,"rwx");break;
                        default:
                                strcat(t_mode,"???");break;
                }
        }

	return (t_mode);
}

////////////////////////////////////////////////////////////////////////////////
// This routine will send an event to the user-selected output device.
////////////////////////////////////////////////////////////////////////////////
int sendevent(char *string)
{
	char stringout[MAX_AUDITREC] = "";

	// NEW audit record format, including host and LinuxAudit identifier string:
	snprintf(stringout, MAX_AUDITREC, "%s\tLinuxAudit\t%s", hostid,
		 string);

	if (AuditDestination & AUDIT_TO_STDOUT) {
		// Send to STDOUT.
		printf("%s\n", stringout);
	}

	if (AuditDestination & AUDIT_TO_FILE) {
		// Write to file
		if (AuditFile != (FILE *) NULL) {
			fprintf(AuditFile, "%s\n", stringout);
			fflush(AuditFile);	// expensive, huh?
		}
	}

	if (AuditDestination & AUDIT_TO_NETWORK) {
                if(AuditDestination & AUDIT_TO_SYSLOG) {
                        // Do Syslog stuff here.
                        time_t currenttime;
                        struct tm *newtime;
                        char CurrentDate[16]="";
                                                                                
                        time(&currenttime);
                        newtime=localtime(&currenttime);
                        syslogdate(CurrentDate,newtime);
                                                                                
			snprintf(stringout, MAX_AUDITREC, "<%d> %s %s LinuxAudit\t%s", 
				SyslogDestination,CurrentDate,hostid, string);
                }
		if (AuditSocket) {
			SendToSocket(stringout);
		}
	}

	return (0);
}

// simple function to do username lookups on uid and effective uid's.
// this function caches the last two results so that we don't have to
// do additional lookups.  we also cache two results so that we can
// process the results from two successive calls to passname (in a
// large va_eventcat call, for instance) without getting in our own
// way.
char *passname(int uid)
{
	static int uidcache = -1;
	static char cachestr[MAX_USERNAME] = "";

	static int uidcache2 = -1;
	static char cachestr2[MAX_USERNAME] = "";

	static char *unknown = "unknown";

	static int lastitem = 0;	// flips between 0 and 1

	struct passwd *passent;

	if (uid == uidcache) {
		return cachestr;
	} else if (uid == uidcache2) {
		return cachestr2;
	}

	// didn't find it in the cache
	passent = getpwuid((uid_t) uid);

	if ((struct passwd *) NULL == passent) {
		return unknown;
	}

	// but we did find it from getpwuid(), so let's copy the result into
	// an internal cache buffer and return a pointer

	if (lastitem) {
		strncpy(cachestr, passent->pw_name, MAX_USERNAME);
		uidcache = uid;
		lastitem = 0;
		return cachestr;
	} else {
		strncpy(cachestr2, passent->pw_name, MAX_USERNAME);
		uidcache2 = uid;
		lastitem = 1;
		return cachestr2;
	}
}

// simple function to do groupname lookups on gid and effective gid's.
// this function caches the last two results so that we don't have to
// do additional lookups.  we also cache two results so that we can
// process the results from two successive calls to groupname (in a
// large va_eventcat call, for instance) without getting in our own
// way.

char *groupname(int gid)
{
	static int gidcache = -1;
	static int gidcache2 = -1;

	static char cachestr[MAX_USERNAME] = "";
	static char cachestr2[MAX_USERNAME] = "";

	static char *unknown = "unknown";

	static int lastitem = 0;	// flips between 0 and 1

	struct group *groupent;

	if (gid == gidcache) {
		return cachestr;
	} else if (gid == gidcache2) {
		return cachestr2;
	}
	// didn't find it in the cache

	groupent = getgrgid((uid_t) gid);

	if ((struct group *) NULL == groupent) {
		return unknown;
	}
	// but we did find it from getgrgid(), so let's copy the result into
	// an internal cache buffer and return a pointer

	if (lastitem) {
		strncpy(cachestr, groupent->gr_name, MAX_USERNAME);
		gidcache = gid;
		lastitem = 0;
		return cachestr;
	} else {
		strncpy(cachestr2, groupent->gr_name, MAX_USERNAME);
		gidcache2 = gid;
		lastitem = 1;
		return cachestr2;
	}
}

// If we want to use a config file that uses text-type event names to enable/disable events
int eventname2number(char *eventname)
{
	if (!strcmp(eventname, "open"))
		return AUDIT_open;

	if (!strcmp(eventname, "execve"))
		return AUDIT_execve;

	if (!strcmp(eventname, "exit"))
		return AUDIT_exit;

	if (!strcmp(eventname, "fork"))
		return AUDIT_fork;

	if (!strcmp(eventname, "mkdir"))
		return AUDIT_mkdir;

	if (!strcmp(eventname, "unlink"))
		return AUDIT_unlink;

	if (!strcmp(eventname, "mknod"))
		return AUDIT_mknod;

	if (!strcmp(eventname, "rmdir"))
		return AUDIT_rmdir;

	if (!strcmp(eventname, "chown"))
		return AUDIT_chown;

	if (!strcmp(eventname, "lchown"))
		return AUDIT_chown;

	if (!strcmp(eventname, "chown32"))
		return AUDIT_chown;

	if (!strcmp(eventname, "lchown32"))
		return AUDIT_chown;

	if (!strcmp(eventname, "chmod"))
		return AUDIT_chmod;

	if (!strcmp(eventname, "symlink"))
		return AUDIT_symlink;

	if (!strcmp(eventname, "link"))
		return AUDIT_link;

	if (!strcmp(eventname, "rename"))
		return AUDIT_rename;

	if (!strcmp(eventname, "reboot"))
		return AUDIT_reboot;

	if (!strcmp(eventname, "truncate"))
		return AUDIT_truncate;

	if (!strcmp(eventname, "chroot"))
		return AUDIT_chroot;

	if (!strcmp(eventname, "setuid"))
		return AUDIT_setuid;

	if (!strcmp(eventname, "setreuid"))
		return AUDIT_setreuid;

	if (!strcmp(eventname, "setresuid"))
		return AUDIT_setresuid;

	if (!strcmp(eventname, "setgid"))
		return AUDIT_setgid;

	if (!strcmp(eventname, "setregid"))
		return AUDIT_setregid;

	if (!strcmp(eventname, "setresgid"))
		return AUDIT_setresgid;

	if (!strcmp(eventname, "connect"))
		return AUDIT_connect;

	if (!strcmp(eventname, "accept"))
		return AUDIT_accept;

	if (!strcmp(eventname, "create_module"))
		return AUDIT_create_module;

	if (!strcmp(eventname, "delete_module"))
		return AUDIT_delete_module;

	if (!strcmp(eventname, "mount"))
		return AUDIT_mount;

	if (!strcmp(eventname, "umount"))
		return AUDIT_umount;

	// We should not get here.
	return (-1);
}

// If we want to use a config file that uses text-type event names to enable/disable events
char *eventnumber2name(int number, char *eventname, int buffer)
{
	switch (number) {
	case AUDIT_open:
		strncpy(eventname, "open", buffer);
		break;
	case AUDIT_execve:
		strncpy(eventname, "execve", buffer);
		break;
	case AUDIT_exit:
		strncpy(eventname, "exit", buffer);
		break;
	case AUDIT_fork:
		strncpy(eventname, "fork", buffer);
		break;
	case AUDIT_mkdir:
		strncpy(eventname, "mkdir", buffer);
		break;
	case AUDIT_unlink:
		strncpy(eventname, "unlink", buffer);
		break;
	case AUDIT_mknod:
		strncpy(eventname, "mknod", buffer);
		break;
	case AUDIT_rmdir:
		strncpy(eventname, "rmdir", buffer);
		break;
	case AUDIT_chown:
		strncpy(eventname, "chown", buffer);
		break;
	case AUDIT_chmod:
		strncpy(eventname, "chmod", buffer);
		break;
	case AUDIT_symlink:
		strncpy(eventname, "symlink", buffer);
		break;
	case AUDIT_link:
		strncpy(eventname, "link", buffer);
		break;
	case AUDIT_rename:
		strncpy(eventname, "rename", buffer);
		break;
	case AUDIT_reboot:
		strncpy(eventname, "reboot", buffer);
		break;
	case AUDIT_truncate:
		strncpy(eventname, "truncate", buffer);
		break;
	case AUDIT_chroot:
		strncpy(eventname, "chroot", buffer);
		break;
	case AUDIT_setuid:
		strncpy(eventname, "setuid", buffer);
		break;
	case AUDIT_setreuid:
		strncpy(eventname, "setreuid", buffer);
		break;
	case AUDIT_setresuid:
		strncpy(eventname, "setresuid", buffer);
		break;
	case AUDIT_setgid:
		strncpy(eventname, "setgid", buffer);
		break;
	case AUDIT_setregid:
		strncpy(eventname, "setregid", buffer);
		break;
	case AUDIT_setresgid:
		strncpy(eventname, "setresgid", buffer);
		break;
	case AUDIT_ftruncate:
		strncpy(eventname, "ftruncate", buffer);
		break;
	case AUDIT_connect:
		strncpy(eventname, "connect", buffer);
		break;
	case AUDIT_accept:
		strncpy(eventname, "accept", buffer);
		break;
	case AUDIT_create_module:
		strncpy(eventname, "create_module", buffer);
		break;
	case AUDIT_delete_module:
		strncpy(eventname, "delete_module", buffer);
		break;
	case AUDIT_mount:
		strncpy(eventname, "mount", buffer);
		break;
	case AUDIT_umount:
		strncpy(eventname, "umount", buffer);
		break;
	default:
		strncpy(eventname, "", buffer);
		break;
	}
	return (eventname);
}

char *resolved_path(char *pwd, char *path, char *resolved)
{
	char temppath[PATH_MAX + PATH_MAX + 1];
	char *tresolved;

	if (path[0] != '/') {
		snprintf(temppath,sizeof(temppath),"%s/%s",pwd,path);
		if ((tresolved =
		     realpath(temppath, resolved)) == (char *) NULL) {
			// Cannot resolve the path - do what we can to get rid of
			// . and .. out of the path.
			resolved_path2(temppath, resolved);
			return (resolved);
		} else {
			return (tresolved);
		}
	} else {
		// Path is already absolute. return it.
		if (realpath(path, resolved) == (char *) NULL) {
			// Could not resolve for some reason.
			resolved_path2(path, resolved);
			return (resolved);
		} else {
			return (resolved);
		}
	}
}


int resolved_path2(char *path, char *resolved_path)
{
	char *new_path = resolved_path;
	char *max_path = resolved_path + PATH_MAX - 1;

	if (!path || !resolved_path) {
		return (0);
	}
	// Keep the first slash
	if (*path == '/') {
		*new_path++ = *path++;
	}

	while (*path != '\0' && (new_path < max_path)) {
		// Ignore stray "/".
		if (*path == '/') {
			path++;
			continue;
		}
		if (*path == '.') {
			// Ignore ".".
			if (path[1] == '\0' || path[1] == '/') {
				path++;
				continue;
			}
			if (path[1] == '.') {
				if (path[2] == '\0' || path[2] == '/') {
					path += 2;
					// Ignore ".." at root.
					if (new_path == resolved_path + 1)
						continue;
					// Handle ".." by backing up.
					while ((--new_path)[-1] != '/');
					continue;
				}
			}
		}
		// Safely copy the next pathname component.
		while (*path != '\0' && *path != '/') {
			if (new_path >= max_path) {
				return (0);
			}
			*new_path++ = *path++;
		}
		if (*path == '/') {
			*new_path++ = '/';
		}
	}
	return(0);
}

// function to process objectives from the objective list.
//
// NOTE: expects start to be pointing to an objective Node already
// prevalidated by a call to FindNextObjective.  If you call this
// function with a start node that you have not previously matched
// against the current event, it'll proceed to log it, so beware.

void
ProcessObjectives(Node * start,
		  header_token * header, const char *searchterm,
		  const char *options, const char *objectivestr,
		  const char *eventdata, const char *timestr,
		  int successfail)
{
	Node *nodematch;
	char *username;
	char objectivedata[MAX_AUDITREC];

	// remember this passname() call will wipe out the passname() result
	// before last, if it's not one of the last two user id numbers
	// we've seen
	username = passname(header->euser_id);

	for (nodematch = start; nodematch;
	     nodematch = FindNextObjective(nodematch->next,
					   header->event_id, username,
					   searchterm, options,
					   header->returncode,
					   successfail)) {
		objectivedata[0] = '\0';

		va_eventcat(objectivedata, 9,
			    "objective",
			    IDELIMITER,
			    criticalitystring(nodematch->criticality),
			    IDELIMITER,
			    timestr, IDELIMITER, objectivestr, DELIMITER,
			    eventdata);

		sendevent(objectivedata);
	}
}

// function to find a matching objective from our objective list.
// start should be the node to start the search at.. generally either
// 'head' or the last known good node->next.

Node *FindNextObjective(Node * start, int eventid, const char *username,
			const char *searchterm, const char *options,
			int returncode, int successfail)
{
	int matched;
	Node *currentnode;

	for (currentnode = start; currentnode;
	     currentnode = currentnode->next) {
		if (eventid != currentnode->event_number
		    && currentnode->event_number != AUDIT_ALL) {
			continue;
		}

		// DeMorgan's transform of old CheckObjective pre-check, negated

		// If the 'returnflag' is set, match it against the
		// 'successfail' variable (reverse if NOTRETURNFLAG is set).
		//
		// Otherwise, match the actual returncode against the returncode
		// variable (reverse if NOTRETURNFLAG is set).

		if ((!currentnode->returnflag ||
		     (currentnode->returnflag != RETURNCODE_ANY &&
		      ((currentnode->notreturnflag
			|| successfail != currentnode->returnflag)
		       && (currentnode->notreturnflag != 1
			   || successfail == currentnode->returnflag))))
		    && (currentnode->returnflag
			||
			((currentnode->notreturnflag
			  || returncode != currentnode->returncode)
			 && (currentnode->notreturnflag != 1
			     || returncode == currentnode->returncode)))) {
			continue;
		}

		matched =
		    (regexec
		     (&currentnode->username_regexp, username, (size_t) 0,
		      NULL, 0) == 0);

		// if we are trying to exclude matches and we matched, skip this
		// objective.. ditto if we are not trying to exclude and we did
		// not match

		if (currentnode->excludeflag == matched) {
			continue;
		}


		// Add in the match include/exclude capability
		if(currentnode->excludematchflag ^ regexec(&currentnode->path_regexp, searchterm, (size_t) 0, NULL, 0)) {
			continue;
		}

		if(!regexec(&currentnode->options_regexp, options,
				(size_t) 0, NULL, 0)) {
			break;	// found one
		}
	}
	return currentnode;
}

// Open, read and act on the configuration file.
// Return AUDIT_BY_OBJECTIVE or AUDIT_BY_EVENT or -1 for error.
int read_config_file()
{
	FILE *configfile = (FILE *) NULL;
	char inputbuffer[MAX_AUDIT_CONFIG_LINE];	// Should be enough for most config lines.
	// Would live to use gchar here instead, but need to keep this simple.
	// Config file header.
	int headertype = 0;
	int audittype = AUDIT_BY_OBJECTIVE;	// Default.

	// Clear the audit destination, and other default variables.
	AuditDestination = 0;

	remote_allow=0;
	remote_webport=6161;
	strncpy(remote_restrictip,"",sizeof(remote_restrictip));
	strncpy(remote_password,"",sizeof(remote_password));

	configfile = fopen(CONFIG_FILENAME, "r");

	if (configfile == (FILE *) NULL) {
		return (AUDIT_BY_OBJECTIVE);
	}

	while (fgets(inputbuffer, MAX_AUDIT_CONFIG_LINE, configfile)) {
		// Kill whitespace from start and end of line.
		trim(inputbuffer);

		if (iscomment(inputbuffer)) {
			continue;
		}
		// Is this line a header?  if so, parse and continue

		if (isheader(inputbuffer)) {
			headertype = getheader(inputbuffer);
			continue;
		}

		if (headertype == CONFIG_AUDITTYPE) {
			// Audit by objective, or not?

			if (isaudittype(inputbuffer)) {
				audittype = getaudittype(inputbuffer);
			}
		} else if (headertype == CONFIG_OBJECTIVES) {
			// Save off enough space to store the data we need.
			char event[MAX_AUDIT_CONFIG_LINE];
			char user[MAX_AUDIT_CONFIG_LINE];
			char path[MAX_AUDIT_CONFIG_LINE];
			char event2[MAX_AUDIT_CONFIG_LINE],
			    options[MAX_OPTIONS] = "";
			char *eventpointer = NULL;
			char *eventpointer2 = NULL;
			int criticality;
			int eventnumber;
			int returncode;
			int returnflag;
			int excludeflag = 0;
			int excludematchflag = 0;
			int notreturnflag = 0;

			if ((criticality =
			     splitobjective(inputbuffer,
					    event, user,
					    path,
					    &returncode,
					    &returnflag,
					    &excludeflag,
					    &excludematchflag,
					    &notreturnflag)) > -1) {
				// add the objective to the linked list.
				trimallwhitespace(event);

				// HERE: Split event into comma separated list of events                                
				eventpointer = event;
				eventpointer2 = event;

				// While there are no more commas
				while (eventpointer2) {
					eventpointer2 =
					    strstr(eventpointer, ",");
					if (eventpointer2 == (char *) NULL) {
						// No commas left. Just copy to the end of the line.
						strncpy(event2,
							eventpointer,
							MAX_AUDIT_CONFIG_LINE);
					} else {
						strncpy
						    (event2,
						     eventpointer,
						     (eventpointer2 -
						      eventpointer));
						// Make sure we have a null on the end of the line.
						event2[eventpointer2 -
						       eventpointer] =
						    '\0';
					}
					if (eventpointer2) {
						// Skip the comma
						eventpointer =
						    eventpointer2 + 1;
					}
					// Clear out any options strings
					strncpy(options,"",sizeof(options));
					if (regmatch(event2, "\\(.*\\)$")) {
						char *position;

						// Bonus, we have something. Grab it, add to options.
						// First, trim off the last bracket.
						event2[strlen(event2) -
						       1] = '\0';
						position =
						    strstr(event2, "(");

						strncpy(options,
							position + 1,
							MAX_OPTIONS);
						// truncate the event string at the bracket.
						*position = '\0';
					}

					if ((eventnumber =
					     eventname2number(event2)) >
					    -1) {
						AddToList(eventnumber,
							  user, path,
							  criticality,
							  options,
							  excludeflag,
							  excludematchflag,
							  returncode,
							  returnflag,
							  notreturnflag);

						// Also, ensure this event is turned on if we are auditing by objective.
						// Note that this may not have been in the config file yet...
						// Best to assume objective auditing is NOT on though.
						if (audittype ==
						    AUDIT_BY_OBJECTIVE) {
							turn_event_on
							    (eventnumber);
						}
					} else {
						if (!strcmp(event, "*")) {
							AddToList
							    (AUDIT_ALL,
							     user,
							     path,
							     criticality,
							     "",
							     excludeflag,
							     excludematchflag,
							     returncode,
							     returnflag,
							     notreturnflag);
						}
					}
				}
			} else {
				fprintf(stderr,
					"WARNING: Cannot process objective - please ensure configuration line\n\tcontains valid criticality, event, criticality, return, user and match elements\n\t%s\n",
					inputbuffer);
			}
		} else if (headertype == CONFIG_EVENTS) {
			int eventnumber = 0;

			// Do not turn events on if we are auditing by objectives at the moment.
			// Note that this will stop event="*" from working on any events that have not been explicitly defined
			// within an objective. However, it is appropriate from a usability perspective.
			if (audittype != AUDIT_BY_OBJECTIVE) {
				eventnumber = geteventswitch(inputbuffer);
				if (eventnumber > -1) {
					// Turn it on.
					turn_event_on(eventnumber);
				}
			}
		} else if (headertype == CONFIG_OUTPUT) {
			if (isfilename(inputbuffer)) {
				if (!open_audit_output(inputbuffer)) {
					fprintf(stderr,
						"WARNING: Could not open the file specified in the audit configuration: %s - Auditing to STDOUT\n",
						inputbuffer);
				}
			} else if (isnetwork(inputbuffer)) {
				if (!open_audit_network(inputbuffer)) {
					fprintf(stderr,
						"WARNING: Could not open the network specified in the audit configuration: %s - Auditing to STDOUT\n",
						inputbuffer);
				}
			} else if(issyslog(inputbuffer)) {
				SyslogDestination=get_syslog_dest(inputbuffer);
                                if(!SyslogDestination) {
                                	SyslogDestination=13;
                                        fprintf(stderr,"WARNING: Could not establish the correct syslog destination from the audit configuration: %s - Sending to user.notice\n",inputbuffer);
                                }
			}
		} else if (headertype == CONFIG_HOSTID) {
			gethostident(inputbuffer, hostid, MAX_HOSTID);
		} else if(headertype == CONFIG_REMOTE) {
			// Grab the remote control stuff here
			if(regmatchi(inputbuffer,"^allow=")) {
				if(regmatchi(inputbuffer,"=1")) {
					remote_allow=1;
				}
			} else if(regmatchi(inputbuffer,"^listen_port=")) {
				remote_webport=getport(inputbuffer);
			} else if(regmatchi(inputbuffer,"^restrict_ip=")) {
				if(!getconfstring(inputbuffer,remote_restrictip,sizeof(remote_restrictip))) {
					strncpy(remote_restrictip,"",sizeof(remote_restrictip));
				}
			} else if(regmatchi(inputbuffer,"^accesskey=")) {
				if(!getconfstring(inputbuffer,remote_password,sizeof(remote_password))) {
					strncpy(remote_password,"",sizeof(remote_password));
				}
			}
		} else {
			fprintf(stderr,
				"WARNING: Configuration file line does not fit in any recognised header.\n\t%s\n",
				inputbuffer);
		}
	}

	fclose((FILE *) configfile);
	return (audittype);
}

int iscomment(char *line)
{
	// Verify that there is something to check.
	if (line == (char *) NULL)
		return (1);

	// And that there is some data within.
	if (strlen(line) == 0 || strlen(line) > MAX_AUDIT_CONFIG_LINE)
		return (1);

	// If the first non-whitespace character is a hash, this is a comment line.
	while (*line) {
		// Space or tab or newline / formfeed
		if (*line == ' ' || *line == 9 || *line == 10
		    || *line == 12) {
			line++;
		} else if (*line == '#') {
			return (1);
		} else {
			// Ahh. A non-whitespace, non hash character.
			return (0);
		}
	}

	// If we are here, then the whole line must have been whitespace
	return (1);
}

int isheader(char *string)
{
	if (string[0] == '[' && string[strlen(string) - 1] == ']') {
		return (1);
	}
	return (0);
}

int isaudittype(char *string)
{
	if (regmatch(string, "^type=")) {
		return (1);
	}
	return (0);
}

int isfilename(char *string)
{
	if (regmatch(string, "^file=")) {
		return (1);
	} else {
		return (0);
	}
}

// Network output
int isnetwork(char *string)
{
	if (regmatch(string, "^network=")) {
		return (1);
	} else {
		return (0);
	}
}
                                                                                
int issyslog(char *string)
{
        if(regmatch(string,"^syslog=")) {
                return(1);
        }
        return(0);
}


int open_audit_output(char *string)
{
	string += strlen("file=");
	if (strlen(string)) {
		if (regmatchi(string, "stdout")) {
			AuditDestination |= AUDIT_TO_STDOUT;
			return (1);
		}
		AuditFile = fopen(string, "a");
		if (AuditFile != (FILE *) NULL) {
			AuditDestination |= AUDIT_TO_FILE;
			return (1);
		} else {
			AuditDestination |= AUDIT_TO_STDOUT;
			return (0);
		}
	}

	return (0);
}

// For the moment, require the web server to do the work of converting
// destinations to numerics.
int get_syslog_dest(char *string)
{
        int destination;
        string+=strlen("syslog=");
        destination=atoi(string);
        if(destination > 0) {
                return(destination);
        }
        return(0);
}


// Open the network - destination host and port
int open_audit_network(char *string)
{
	char desthost[MAX_HOSTID];
	int port = 6161;	// Default port
	char *pos;
	struct hostent *hp;

	// Initialise desthost.
	strncpy(desthost, "localhost", MAX_HOSTID);

	string += strlen("network=");
	if (strlen(string)) {
		if (regmatchi(string, "stdout")) {
			AuditDestination |= AUDIT_TO_STDOUT;
			return (1);
		}

		AuditSocket = socket(AF_INET, SOCK_DGRAM, 0);
		if (AuditSocket < 0) {
			fprintf(stderr,
				"Cannot open network socket - auditing to STDOUT\n");
			AuditDestination |= AUDIT_TO_STDOUT;
			return (0);
		}

		pos = strstr(string, ":");
		if (pos) {
			if ((pos - string) < MAX_HOSTID) {
				strncpy(desthost, string, (pos - string));
				desthost[pos - string] = '\0';
			} else {
				strncpy(desthost, string, MAX_HOSTID);
				desthost[MAX_HOSTID - 1] = '\0';
			}
			pos++;
			// As long as the user has a number there.
			if (strlen(pos)) {
				port = atoi(pos);
			}
		} else {
			// No colon. Assume the default port to be 6161
			strncpy(desthost, string, MAX_HOSTID);
		}

		// Just in case the user has entered a blank field
		if (strlen(desthost) == 0) {
			strncpy(desthost, "localhost", MAX_HOSTID);
		}

		hp = gethostbyname(desthost);
		if (hp == 0) {
			fprintf(stderr,
				"Cannot resolve host %s - auditing to STDOUT\n",
				hostid);
			close(AuditSocket);
			AuditDestination |= AUDIT_TO_STDOUT;
			return (0);
		}

		bcopy(hp->h_addr, &AuditSocketName.sin_addr, hp->h_length);
		AuditSocketName.sin_family = AF_INET;
		AuditSocketName.sin_port = htons(port);

		AuditDestination |= AUDIT_TO_NETWORK;
		return (1);
	}

	AuditDestination |= AUDIT_TO_STDOUT;
	return (0);
}


int getheader(char *string)
{
	char *stringp = string;

	if (regmatchi(stringp, "^\\[audittype\\]$")) {
		return (CONFIG_AUDITTYPE);
	} else if (regmatchi(stringp, "^\\[objectives\\]$")) {
		return (CONFIG_OBJECTIVES);
	} else if (regmatchi(stringp, "^\\[events\\]$")) {
		return (CONFIG_EVENTS);
	} else if (regmatchi(stringp, "^\\[file\\]$")
		   || regmatchi(stringp, "^\\[output\\]$")) {
		return (CONFIG_OUTPUT);
	} else if (regmatchi(stringp, "^\\[hostid\\]$")) {
		return (CONFIG_HOSTID);
	} else if (regmatchi(stringp, "^\\[remote\\]$")) {
		return (CONFIG_REMOTE);
	}

	fprintf(stderr, "Unknown header in configuration file: %s\n",
		stringp);
	return (0);
}

int getaudittype(char *string)
{
	char *stringp = string;
	// Get rid of the type= part.
	stringp += strlen("type=");
	if (regmatchi(stringp, "^objective$")) {
		return (AUDIT_BY_OBJECTIVE);
	} else {
		return (AUDIT_BY_EVENT);
	}
}

int geteventswitch(char *string)
{
	char *stringp = string;
	char *endname = string;

	stringp = strstr(string, "=");
	if (stringp != (char *) NULL) {
		endname = stringp;
		stringp++;
		if (!strcmp(stringp, "1")) {
			// Set this event.
			// Grab the event name. Insert a null into the string in place of the = character.
			*endname = '\0';
			return (eventname2number(string));
		}
	} else {
		return (-1);
	}
	return (-1);
}

// Return the host identifier
char *gethostident(char *string, char *host, int length)
{
	char *stringp = string;

	stringp = strstr(string, "=");
	if (stringp != (char *) NULL) {
		stringp++;
		if (strlen(stringp)) {
			strncpy(host, stringp, length - 1);
		} else {
			return ((char *) NULL);
		}
	} else {
		return ((char *) NULL);
	}
	return ((char *) host);
}

// Return a string that contains the criticality
char *criticalitystring(int criticality)
{
	switch (criticality) {
	case CRITICALITY_CRITICAL:
		return "critical";

	case CRITICALITY_PRIORITY:
		return "priority";

	case CRITICALITY_WARNING:
		return "warning";

	case CRITICALITY_INFO:
		return "information";

	default:
		return "clear";
	}
}

// Remove start / end whitespace, including the newline
void trim(char *string)
{
	char *pointer;
	char *walker;
	char *header;
	int blankmarker=1;

	// Verify that there is something to check.
	if (string == (char *) NULL)
		return;

	// And that there is some data within.
	if (strlen(string) == 0)
		return;

	// Start from the end, work backwards.
	pointer = string;
	pointer += strlen(string) - 1;

	while ((*pointer == ' ' || *pointer == 9 || *pointer == 10
		|| *pointer == 12) && pointer >= string) {
		*pointer = '\0';
		pointer--;
	}

	// Are we back at the start of the string? If so, this line must be null.
	if (pointer == string)
		return;

	// Pointer is now at the last non-whitespace character of the string.
	walker = string;
	header = string;
	while(walker <= pointer) {
		if((*walker == ' ' || *walker == 9 || *walker == 10
		   || *walker == 12) && blankmarker) {
			walker++;
		} else {
			blankmarker=0;

			if(*header != *walker) {
				*header=*walker;
			}
			header++;
			walker++;
		}
	}
	*header='\0';
}

void trimallwhitespace(char *string)
{
	char *readpointer;
	char *writepointer;

	// Verify that there is something to check.
	if (string == (char *) NULL)
		return;

	// And that there is some data within.
	if (strlen(string) == 0)
		return;

	// Start from the end, work backwards.
	readpointer = string;
	writepointer = string;
	while (readpointer < (string + strlen(string))) {
		if (*readpointer == ' ' || *readpointer == 9
		    || *readpointer == 10 || *readpointer == 12) {
			readpointer++;
		} else {
			if (writepointer != readpointer) {
				*writepointer = *readpointer;
			}
			readpointer++;
			writepointer++;
		}

	}
}

// Strip tabs and newlines out of a string.
char *strip(char *string)
{
	char *stringp = string;

	while (*stringp != '\0') {
		if (*stringp == '\t' || *stringp == '\n') {
			*stringp = ' ';
		}
		stringp++;
	}
	return (string);
}

// Take an identified objective line
// NOTE: string, event, user and match must all be the same buffer size!
int
splitobjective(char *string, char *event, char *user, char *match,
	       int *returncode, int *returnflag, int *excludeflag,
	       int *excludematchflag, int *notreturnflag)
{
	char *startevent, *startuser, *startmatch, *startcrit,
	    *startreturn;
	char *stringpointer, *endstring;
	char *eventp, *userp, *matchp, *critp;
	char criticality[MAX_AUDIT_CONFIG_LINE];
	int crit = 0;

	*returnflag = RETURNCODE_ANY;

	// Do some basic sanity checks.
	if (string == (char *) NULL || event == (char *) NULL
	    || user == (char *) NULL || match == (char *) NULL) {
		return (-1);
	}

	startcrit = strstr(string, "criticality=");
	startevent = strstr(string, "event=");
	startreturn = strstr(string, "return=");
	if (startreturn == (char *) NULL) {
		startreturn = strstr(string, "return!=");
		if (startreturn != (char *) NULL) {
			// NOT return rather than return.
			*notreturnflag = 1;
		}
	}

	startmatch = strstr(string, "match=");
	if(startmatch == (char *)NULL) {
		startmatch=strstr(string,"match!=");
		if(startmatch!=(char *)NULL) {
			// EXCLUDE rather than include.
			*excludematchflag=1;
		}
	}

	startuser = strstr(string, "user=");
	if (startuser == (char *) NULL) {
		startuser = strstr(string, "user!=");
		if (startuser != (char *) NULL) {
			// EXCLUDE USERS rather than include.
			*excludeflag = 1;
		}
	}
	startmatch = strstr(string, "match=");
	if(!startmatch) {
		startmatch = strstr(string, "match!=");
	}

	// string pointers for iteration.
	stringpointer = string;
	eventp = event;
	userp = user;
	matchp = match;
	critp = criticality;

	// Pointer to the last character in the string.
	endstring = &string[strlen(string)];

	if (startevent == (char *) NULL || startuser == (char *) NULL ||
	    startmatch == (char *) NULL || startcrit == (char *) NULL
	    || startreturn == (char *) NULL) {
		// Problem, this line is malformed. We really cannot proceed with this line.
		fprintf(stderr,
			"The following line does not contain the criticality, event, return, user and match elements: %s",
			string);
		return (-1);
	}
	// Start with the event
	stringpointer = startcrit + strlen("criticality=");
	while (*stringpointer
	       && (stringpointer != startcrit
		   && stringpointer != startreturn
		   && stringpointer != startuser
		   && stringpointer != startmatch
		   && stringpointer != endstring)) {
		*critp = *stringpointer;
		critp++;
		stringpointer++;
	}
	*critp = '\0';
	trim(criticality);
	crit = atoi(criticality);
	if (crit < CRITICALITY_CLEAR)
		crit = CRITICALITY_CLEAR;
	if (crit > CRITICALITY_CRITICAL)
		crit = CRITICALITY_CRITICAL;

	// Start with the event
	stringpointer = startevent + strlen("event=");
	while (*stringpointer
	       && (stringpointer != startcrit
		   && stringpointer != startreturn
		   && stringpointer != startuser
		   && stringpointer != startmatch
		   && stringpointer != endstring)) {
		*eventp = *stringpointer;
		eventp++;
		stringpointer++;
	}
	*eventp = '\0';
	// Remove extra whitespace at start and end.
	trim(event);

	stringpointer = startreturn + strlen("return=");
	// Just in case it was a !=
	if (*stringpointer == '=') {
		stringpointer++;
	}

	if (regmatchi(stringpointer, "^success")) {
		*returnflag = RETURNCODE_SUCCESS;
		*returncode = 0;
	} else if (regmatchi(stringpointer, "^failure")) {
		*returnflag = RETURNCODE_FAILURE;
		*returncode = 0;
	} else {
		if (*stringpointer == '*'
		    || (*stringpointer < '0' || *stringpointer > '9')) {
			*returnflag = RETURNCODE_ANY;
			// Notreturnflag does not make sense for ANY
			*notreturnflag = 0;
			*returncode = 0;
		} else {
			*returncode = atoi(stringpointer);
			*returnflag = 0;
		}
	}

	if (*excludeflag) {
		stringpointer = startuser + strlen("user!=");
	} else {
		stringpointer = startuser + strlen("user=");
	}

	while (*stringpointer
	       && (stringpointer != startcrit
		   && stringpointer != startreturn
		   && stringpointer != startmatch
		   && stringpointer != startevent
		   && stringpointer != endstring)) {
		*userp = *stringpointer;
		stringpointer++;
		userp++;
	}
	*userp = '\0';
	// Remove extra whitespace at start and end.
	trim(user);

	if(*excludematchflag) {
		stringpointer=startmatch+strlen("match!=");
	} else {
		stringpointer=startmatch+strlen("match=");
	}
	while (*stringpointer
	       && (stringpointer != startcrit
		   && stringpointer != startreturn
		   && stringpointer != startevent
		   && stringpointer != startuser
		   && stringpointer != endstring)) {
		*matchp = *stringpointer;
		stringpointer++;
		matchp++;
	}
	*matchp = '\0';
	// Remove extra whitespace at start and end.
	trim(match);

	// Return the criticality value.
	return (crit);
}

// Return the host identifier
char * getconfstring(char *string, char *file,int length)
{
        char *stringp=string;
                                                                                
        stringp=strstr(string,"=");
        if(stringp != (char *) NULL) {
                stringp++;
                if(strlen(stringp)) {
                        strncpy(file,stringp,length-1);
                } else {
                        return((char *)NULL);
                }
        } else {
                return((char *)NULL);
        }
        return((char *)file);
}

int getport(char *string)
{
        char *stringp=string;
        char strPort[10];
                                                                                
        stringp=strstr(string,"=");
                                                                                
        if(stringp != (char *) NULL) {
                stringp++;
                if(strlen(stringp)) {
                        strncpy(strPort,stringp,10);
                        return(atoi((char *)strPort));
                } else {
                        return(0);
                }
        } else {
                return(0);
        }
}

void syslogdate(char *sdate, struct tm *cdate)
{
        char Month[4];
        char Date[3];
        char Hour[3];
        char Min[3];
        char Sec[3];
                                                                                
        if(!sdate || !cdate) return;
                                                                                
        switch (cdate->tm_mon) {
                case 0: strcpy(Month,"Jan"); break;
                case 1: strcpy(Month,"Feb"); break;
                case 2: strcpy(Month,"Mar"); break;
                case 3: strcpy(Month,"Apr"); break;
                case 4: strcpy(Month,"May"); break;
                case 5: strcpy(Month,"Jun"); break;
                case 6: strcpy(Month,"Jul"); break;
                case 7: strcpy(Month,"Aug"); break;
                case 8: strcpy(Month,"Sep"); break;
                case 9: strcpy(Month,"Oct"); break;
                case 10: strcpy(Month,"Nov"); break;
                default: strcpy(Month,"Dec"); break;
        }
                                                                                
        if(cdate->tm_mday<10) {
                snprintf(Date,3," %d%c",cdate->tm_mday,0);
        } else {
                snprintf(Date,3,"%d%c",cdate->tm_mday,0);
        }
                                                                                
        if(cdate->tm_hour<10) {
                snprintf(Hour,3,"0%d%c",cdate->tm_hour,0);
        } else {
                snprintf(Hour,3,"%d%c",cdate->tm_hour,0);
        }
                                                                                
        if(cdate->tm_min<10) {
                snprintf(Min,3,"0%d%c",cdate->tm_min,0);
        } else {
                snprintf(Min,3,"%d%c",cdate->tm_min,0);
        }
                                                                                
        if(cdate->tm_sec<10) {
                snprintf(Sec,3,"0%d%c",cdate->tm_sec,0);
        } else {
                snprintf(Sec,3,"%d%c",cdate->tm_sec,0);
        }
                                                                                
        snprintf(sdate,16,"%s %s %s:%s:%s%c",Month,Date,Hour,Min,Sec,0);
}

// Match string against an extended regular expression.
// Return 1 for match, 0 for no-match or error.
int regmatch(const char *string, const char *pattern)
{
	int status;
	regex_t re;

	// Two quick checks.
	if (pattern[0] == '*' && pattern[1] == '\0') {
		return (1);
	}
	if (pattern[0] == '.' && pattern[1] == '*' && pattern[2] == '\0') {
		return (1);
	}

	if (regcomp(&re, pattern, REG_EXTENDED | REG_NOSUB) != 0) {
		return (0);
	}
	status = regexec(&re, string, (size_t) 0, NULL, 0);
	regfree(&re);

	if (status != 0) {
		return (0);
	}
	return (1);
}

// Match string against an extended regular expression. Case insensitive.
// Return 1 for match, 0 for no-match or error.
int regmatchi(const char *string, const char *pattern)
{
	int status;
	regex_t re;

	// Two quick checks.
	if (pattern[0] == '*' && pattern[1] == '\0') {
		return (1);
	}

	if (pattern[0] == '.' && pattern[1] == '*' && pattern[2] == '\0') {
		return (1);
	}

	if (regcomp(&re, pattern, REG_EXTENDED | REG_NOSUB | REG_ICASE) !=
	    0) {
		return (0);
	}

	status = regexec(&re, string, (size_t) 0, NULL, 0);

	regfree(&re);

	if (status != 0) {
		return (0);
	}
	return (1);
}


// Linked List Functions

void CreateLinkedList(void)
{
	head = NULL;
}

Node *AddToList(int eventnumber, char *username, char *path,
		int criticality, char *options, int excludeflag,
		int excludematchflag, int returncode, int returnflag,
		int notreturnflag)
{
	Node *newNode = NULL;
	newNode = (Node *) malloc(sizeof(Node));

	if (newNode == NULL) {
		fprintf(stderr,
			"AddToList(): error in dynamic memory allocation\n\
Could not add a new objective into our linked list. You may be low on memory.\n");
		return ((Node *) NULL);
	}

	newNode->event_number = eventnumber;
	newNode->criticality = criticality;
	newNode->returncode = returncode;
	newNode->returnflag = returnflag;	// success or failure (or zero)
	newNode->notreturnflag = notreturnflag;
	newNode->excludeflag = excludeflag;
	newNode->excludematchflag=excludematchflag;

	if (!strcmp(username, "*")) {
		strcpy(username, ".*");
	}

	strcpy(newNode->username, username);

	if (regcomp
	    (&newNode->username_regexp, newNode->username,
	     REG_EXTENDED | REG_NOSUB) != 0) {
		fprintf(stderr,
			"AddToList(): pattern error in username\n%s\n",
			username);

		free(newNode);
		return ((Node *) NULL);
	}

	if (!strcmp(path, "*")) {
		strcpy(path, ".*");
	}

	strcpy(newNode->path, path);

	if (regcomp
	    (&newNode->path_regexp, newNode->path,
	     REG_EXTENDED | REG_NOSUB) != 0) {
		fprintf(stderr, "AddToList(): pattern error in path\n%s\n",
			path);

		regfree(&newNode->username_regexp);
		free(newNode);
		return ((Node *) NULL);
	}

	if (!strcmp(options, "*")) {
		strcpy(options, ".*");
	}

	strncpy(newNode->options, options, MAX_OPTIONS);

	// we do a case-insensitive regexp for options

	if (regcomp
	    (&newNode->options_regexp, newNode->options,
	     REG_EXTENDED | REG_NOSUB | REG_ICASE) != 0) {
		fprintf(stderr,
			"AddToList(): pattern error in options\n%s\n",
			path);

		regfree(&newNode->username_regexp);
		regfree(&newNode->path_regexp);
		free(newNode);
		return ((Node *) NULL);
	}

	if (head == NULL) {
		head = newNode;
		newNode->next = NULL;
	} else {
		newNode->next = head;
		head = newNode;
	}

	return newNode;
}

void RemoveFromListHead(void)
{
	Node *tempPtr;

	if (NULL == head)
		return;

	tempPtr = head;
	head = head->next;

	DestroyNode(tempPtr);
}

void RemoveFromList(Node * node)
{
	Node *tempPtr, *previousPtr;

	if (NULL == node)
		return;

	if (head == node) {
		RemoveFromListHead();
		return;
	}

	tempPtr = head;

	while (NULL != tempPtr) {
		previousPtr = tempPtr;
		tempPtr = tempPtr->next;

		if (tempPtr == node) {
			previousPtr->next = tempPtr->next;
			DestroyNode(tempPtr);
			break;
		}
	}
}

void DestroyNode(Node * node)
{
	regfree(&node->username_regexp);
	regfree(&node->path_regexp);
	regfree(&node->options_regexp);
	free(node);
}

void DestroyList(void)
{
	if (NULL == head) {
		return;
	}

	while (NULL != head) {
		Node *tempPtr = head;
		head = head->next;

		DestroyNode(tempPtr);
	}
}

// Pull out the fully qualified domain name, if possible.
char *getfqdn(char *FQDN)
{
	char hname[MAX_HOSTID];
	struct hostent *hp;

	if (gethostname(hname, MAX_HOSTID)) {
		strncpy(FQDN, "localhost.unknown", MAX_HOSTID);
		return (FQDN);
	}

	strncpy(FQDN, hname, MAX_HOSTID);

	hp = gethostbyname(hname);
	if (hp) {
		while (hp->h_aliases && *hp->h_aliases) {
			if (strlen(*(hp->h_aliases)) > strlen(hname)
			    && !strncmp(hname, *(hp->h_aliases),
					strlen(hname))) {
				strncpy(FQDN, *(hp->h_aliases),
					MAX_HOSTID);
			}
			hp->h_aliases++;
		}
		if (strlen(hp->h_name) > strlen(hname)
		    && !strncmp(hname, hp->h_name, strlen(hname))) {
			strncpy(FQDN, hp->h_name, MAX_HOSTID);
		}
	}

	return (FQDN);
}

int SendToSocket(char *message)
{
	if (AuditSocket) {
		if (sendto(AuditSocket, message,
			   strlen(message), 0,
			   (const struct sockaddr *) &AuditSocketName,
			   sizeof(struct sockaddr)) < 0) {
			puts(message);
			return (1);
		}
	} else {
		puts(message);
		return (1);
	}
	return (0);
}
